from collections import namedtuple
from pathlib import Path
import pandas as pd
import numpy as np
import nilearn as nl
from nilearn import image, masking
from sklearn.cluster import AgglomerativeClustering
from sklearn.feature_extraction.image import grid_to_graph
from scipy import stats
from scipy.spatial import distance
from .utils import select_confounds, idxs_to_flat, idxs_to_zoomed, idxs_to_mask

Cluster = namedtuple("Cluster", "id idxs nvox concentration repts medts")


def clust_within_region(
    mask_path,
    timeseries_path,
    confound_path,
    n_dummy=4,
    t_r=2.5,
    smoothing_fwhm=None,
    confound_selectors=None,
    distance_threshold=0.5,
    adjacency=False,
    metric="precomputed",
    linkage="single",
    clustering=None,
    clustering_kwargs=None,
):
    if confound_selectors is None:
        confound_selectors = ["-motion", "-dummy"]
    if isinstance(mask_path, Path):
        mask_path = mask_path.as_posix()
    if isinstance(timeseries_path, Path):
        timeseries_path = timeseries_path.as_posix()
    if isinstance(confound_path, Path):
        confound_path = confound_path.as_posix()
    if clustering_kwargs is None:
        clustering_kwargs = {}

    sub_mask = nl.image.load_img(mask_path)
    sub_img = nl.image.load_img(timeseries_path)
    cfds = select_confounds(confound_path, confound_selectors)

    rawts = nl.masking.apply_mask(sub_img, sub_mask, smoothing_fwhm=smoothing_fwhm)[
        n_dummy:
    ]
    clnts = nl.signal.clean(
        rawts,
        confounds=cfds[n_dummy:],
        t_r=t_r,
        detrend=True,
        low_pass=0.1,
        high_pass=0.01,
        standardize="psc",
    )
    spr, p = stats.spearmanr(clnts)
    if adjacency:
        connectivity = grid_to_graph(
            sub_mask.shape[0],
            sub_mask.shape[1],
            sub_mask.shape[2],
            mask=sub_mask.get_fdata(),
            return_as=np.ndarray,
        )
    else:
        connectivity = None
    if clustering is None:
        agg = AgglomerativeClustering(
            n_clusters=None,
            metric=metric,
            compute_full_tree=True,
            connectivity=connectivity,
            linkage=linkage,
            distance_threshold=distance_threshold,
        )
    else:
        agg = clustering(**clustering_kwargs)
    # use 1 - spr here to convert the correlation matrix to a distance matrix
    vox_clust_labels = agg.fit_predict(1 - spr)
    cluster_ids, cluster_counts = np.unique(vox_clust_labels, return_counts=True)
    vcl_img = nl.masking.unmask(vox_clust_labels + 1, sub_mask)

    # create cluster objects and append to list
    clusters = []
    for nvox, cid in zip(cluster_counts, cluster_ids):
        idxs = np.where(vox_clust_labels == cid)[0]
        clust_tses = clnts[:, idxs]
        if nvox == 1:
            median_ts = clust_tses.squeeze()
            rep_ts = clust_tses.squeeze()
            concentration = np.nan
            vox_idxs = np.nonzero(vcl_img.get_fdata() == (cid + 1))
            vox_idxs = np.array([(x, y, z, 1) for x, y, z in list(zip(*vox_idxs))])
        else:
            median_ts = np.percentile(clust_tses, 50, axis=1)
            # calculate spearmanr between median ts and each clust_ts
            clust_to_median, _ = stats.spearmanr(median_ts, clust_tses)
            clust_to_median = clust_to_median[0, 1:]
            rep_ts = clust_tses[:, clust_to_median == clust_to_median.max()].squeeze()
            # get a list of voxel indicies of the cluster
            vox_idxs = np.nonzero(vcl_img.get_fdata() == (cid + 1))
            vox_idxs = np.array([(x, y, z, 1) for x, y, z in list(zip(*vox_idxs))])
            vox_locs = np.matmul(vcl_img.affine, vox_idxs.T).T[:, :3]
            assert len(vox_locs) == nvox
            concentration = len(vox_idxs) / distance.pdist(vox_locs).mean()
        cluster = Cluster(cid, vox_idxs[:, :3], nvox, concentration, rep_ts, median_ts)
        clusters.append(cluster)

    return clnts, spr, vox_clust_labels, clusters


def find_target(
    row,
    img_col,
    targout_col,
    distance_threshold=0.5,
    adjacency=False,
    return_df=False,
    write_pkl=False,
    **kwargs,
):
    stimts_cln, stim_spr, stim_vcl, stim_clusters = clust_within_region(
        row.clnstim_mask,
        row[img_col],
        row.confounds,
        distance_threshold=distance_threshold,
        adjacency=adjacency,
        **kwargs,
    )

    refts_cln, ref_spr, ref_vcl, ref_clusters = clust_within_region(
        row.clnref_mask,
        row[img_col],
        row.confounds,
        distance_threshold=distance_threshold,
        adjacency=adjacency,
        **kwargs,
    )

    stim_clust_df = pd.DataFrame(stim_clusters)
    stim_clust_ts = np.array([cc.medts for cc in stim_clusters]).T
    stim_sizes = np.array([cc.nvox for cc in stim_clusters])

    ref_clust_ts = np.array([cc.medts for cc in ref_clusters]).T
    ref_sizes = np.array([cc.nvox for cc in ref_clusters])

    clust_corr, _ = stats.spearmanr(stim_clust_ts, ref_clust_ts)
    try:
        clust_corr = clust_corr[: stim_clust_ts.shape[1], stim_clust_ts.shape[1] :]
        stim_ref_corr = (clust_corr * ref_sizes).sum(1)
    except IndexError:
        stim_ref_corr = (clust_corr * ref_sizes)[0]
    stim_clust_df["net_reference_correlation"] = stim_ref_corr
    for kk, vv in row["entities"].items():
        stim_clust_df[kk] = vv
    if "session" not in stim_clust_df.columns:
        stim_clust_df["session"] = None
    stim_clust_df["adjacency"] = adjacency
    stim_clust_df["distance_threshold"] = distance_threshold
    if write_pkl:
        stim_clust_df.to_pickle(row[targout_col])
    if return_df:
        return stim_clust_df


def rank_clusters(
    targouts_mp, nvox_weight, concentration_weight, net_reference_correlation_weight
):
    targouts_sel = targouts_mp.query("nvox > 50 & net_reference_correlation < 0").copy()
    targouts_sel["flatmask"] = targouts_sel.apply(
        lambda row: idxs_to_flat(row.idxs, row.clnstim_mask), axis=1
    )
    targouts_sel["zoommask"] = targouts_sel.apply(
        lambda row: idxs_to_zoomed(row.idxs, row.clnstim_mask), axis=1
    )
    targouts_sel["nvox_rank"] = targouts_sel.groupby(
        ["run", "session", "subject"], dropna=False
    ).nvox.rank(ascending=False)
    targouts_sel["concentration_rank"] = targouts_sel.groupby(
        ["run", "session", "subject"], dropna=False
    ).concentration.rank(ascending=False)
    targouts_sel["nrc_rank"] = targouts_sel.groupby(
        ["run", "session", "subject"], dropna=False
    ).net_reference_correlation.rank(ascending=True)
    targouts_sel["weighted_nvox_rank"] = targouts_sel.nvox_rank * nvox_weight
    targouts_sel["weighted_concentration_rank"] = (
        targouts_sel.concentration_rank * concentration_weight
    )
    targouts_sel["weighted_nrc_rank"] = (
        targouts_sel.nrc_rank * net_reference_correlation_weight
    )
    targouts_sel["rank_product"] = (
        targouts_sel.weighted_nvox_rank
        * targouts_sel.weighted_concentration_rank
        * targouts_sel.weighted_nrc_rank
    )

    # sorting and using "first" method in rank ensure that there are no ties
    # Consider reordering this based on the weights
    metrics = ["net_reference_correlation", "concentration", "nvox"]
    targouts_sel = targouts_sel.sort_values(metrics).reset_index(drop=True)
    targouts_sel["overall_rank"] = targouts_sel.groupby(
        ["run", "session", "subject"], dropna=False
    ).rank_product.rank(ascending=True, method="first")
    return targouts_sel


def custom_metric(*args, **kwargs):
    if len(args) == 1:
        A = args[0]
        spr, p = stats.spearmanr(A.T)
        return 1 - np.abs(spr)
    elif len(args) == 2:
        A = args[0]
        B = args[1]
        spr, p = stats.spearmanr(A, B)
        return 1 - np.abs(spr)
    else:
        foo = args
        raise NotImplementedError(f"Number of args ({len(args)}) is not implemented.")


def get_clust_image(row):
    clnstim_mask = nl.image.load_img(row.clnstim_mask)
    clnstim_dat = clnstim_mask.get_fdata()

    rowstimmask_dat = idxs_to_mask(row.idxs, row.clnstim_mask)
    rowstimmask = nl.image.new_img_like(
        clnstim_mask, rowstimmask_dat, affine=clnstim_mask.affine, copy_header=True
    )
    # make sure you've got the correct cluster mask
    assert rowstimmask.get_fdata().sum() == row.nvox
    return rowstimmask


def get_vox_ref_corr(
    row,
    distance_threshold,
    adjacency,
    smoothing_fwhm,
    confound_selectors,
    metric,
    linkage,
    t_r,
    n_dummy,
):
    # just using clnstim_mask for now, at some point, it'd be better to use a grey matter mask
    refts_cln, ref_spr, ref_vcl, ref_clusters = clust_within_region(
        row.clnref_mask,
        row.bold_path,
        row.confounds,
        distance_threshold=distance_threshold,
        adjacency=adjacency,
        smoothing_fwhm=smoothing_fwhm,
        confound_selectors=confound_selectors,
        metric=custom_metric,
        linkage=linkage,
        t_r=t_r,
        n_dummy=n_dummy,
    )

    ref_clust_ts = np.array([cc.medts for cc in ref_clusters]).T
    ref_sizes = np.array([cc.nvox for cc in ref_clusters])

    sub_mask = nl.image.load_img(row.clnstim_mask)
    sub_img = nl.image.load_img(row.bold_path)
    cfds = select_confounds(row.confounds, confound_selectors)

    rawts = nl.masking.apply_mask(sub_img, sub_mask, smoothing_fwhm=smoothing_fwhm)[
        n_dummy:
    ]
    clnts = nl.signal.clean(
        rawts,
        confounds=cfds[n_dummy:],
        t_r=t_r,
        detrend=True,
        low_pass=0.1,
        high_pass=0.01,
        standardize="psc",
    )

    clust_corr, _ = stats.spearmanr(clnts, ref_clust_ts)
    try:
        clust_corr = clust_corr[: clnts.shape[1], clnts.shape[1] :]
        vox_ref_corr = (clust_corr * ref_sizes).sum(1)
    except IndexError:
        vox_ref_corr = (clust_corr * ref_sizes)[0]
    vox_ref_corr_img = nl.masking.unmask(vox_ref_corr, sub_mask)
    return vox_ref_corr_img


def get_com_in_mm(row):
    clnstim_mask = nl.image.load_img(row.clnstim_mask)
    try:
        vox_idxs = np.array([list(rr) + [1] for rr in row.idxs])
    except ValueError:
        vox_idxs = np.array([list(rr) + [1] for rr in row.idxs.values[0]])
    vox_locs = np.matmul(clnstim_mask.affine, vox_idxs.T).T[:, :3]
    return vox_locs.mean(0)



import pandas as pd
import numpy as np
import nilearn as nl
from nilearn import image, masking
from niworkflows.interfaces.fixes import (
    FixHeaderApplyTransforms as ApplyTransforms,
)  # pip
from sklearn.feature_extraction.image import grid_to_graph
from scipy.sparse.csgraph import connected_components


# code for transforming things to subject space
def make_path(
    ents,
    updates,
    pattern,
    derivatives_dir,
    build_path,
    check_exist=True,
    check_parent=True,
    mkdir=False,
    make_parent=False,
):
    mp_ents = ents.copy()
    mp_dir = derivatives_dir
    mp_ents.update(updates)
    mp_file = mp_dir / build_path(
        mp_ents, pattern, validate=False, absolute_paths=False
    )
    if check_exist and not mp_file.exists():
        raise FileNotFoundError(mp_file.as_posix())
    elif check_parent and not mp_file.parent.exists():
        raise FileNotFoundError(mp_file.parent.as_posix())
    if mkdir:
        mp_file.mkdir(parents=True, exist_ok=True)
    elif make_parent:
        mp_file.parent.mkdir(parents=True, exist_ok=True)
    return mp_file


def transform_mask_to_t1w(
    row, inmask_col=None, inmask=None, outmask_col="boldmask_path"
):
    if inmask_col is None and inmask is None:
        raise ValueError("one of mask_col or mask must be defined")
    elif inmask_col is not None:
        inmask = row[inmask_col]

    at = ApplyTransforms()
    at.inputs.input_image = inmask
    at.inputs.interpolation = "NearestNeighbor"
    at.inputs.reference_image = row.boldref
    at.inputs.transforms = [row.mnitoT1w]
    at.inputs.output_image = row[outmask_col].as_posix()
    at.inputs.float = True
    _ = at.run()


def clean_mask(
    sub_mask, brain_mask, max_drop_frac=None, clean_mask_path=None, error="raise"
):
    """
    Drop all but the largest of the connected components.
    """
    # mask by subject brain mask
    sub_mask_brain_dat = sub_mask.get_fdata()
    sub_mask_brain_dat[brain_mask.get_fdata() == 0] = 0
    sub_mask_brain = nl.image.new_img_like(
        sub_mask, sub_mask_brain_dat, affine=sub_mask.affine, copy_header=True
    )
    adjacency = grid_to_graph(
        sub_mask_brain.shape[0],
        sub_mask_brain.shape[1],
        sub_mask_brain.shape[2],
        mask=sub_mask_brain_dat,
        return_as=np.ndarray,
    )
    n_connected_components, labels = connected_components(adjacency)
    # add one to labels to make creating a mask easier
    labels = labels + 1
    label_ids, label_counts = np.unique(labels, return_counts=True)
    id_to_keep = label_ids[label_counts == label_counts.max()]

    drop_frac = 1 - (label_counts.max() / len(labels))
    if max_drop_frac is not None:
        if drop_frac > max_drop_frac:
            if error == "raise":
                raise ValueError(
                    f"Keeping only the largest connected component drops\
                {drop_frac:0.3f} of voxels, the maximum dropped fraction is f{max_drop_frac}"
                )
            else:
                print(
                    f"Keeping only the largest connected component drops\
                {drop_frac:0.3f} of voxels, the maximum dropped fraction is f{max_drop_frac}, skipping"
                )
                return None

    print(
        f"Dropping {drop_frac:0.3g} of voxels that are disconnected from the largest component."
    )

    cleaned_mask = nl.masking.unmask(labels, sub_mask_brain)
    cleaned_mask_dat = cleaned_mask.get_fdata()
    cleaned_mask_dat[cleaned_mask_dat != id_to_keep] = 0
    cleaned_mask_dat[cleaned_mask_dat == id_to_keep] = 1
    assert cleaned_mask_dat.sum() > 0
    cleaned_mask = nl.image.new_img_like(
        cleaned_mask,
        data=cleaned_mask_dat,
        affine=cleaned_mask.affine,
        copy_header=True,
    )
    # make sure there's only a single component
    adjacency = grid_to_graph(
        cleaned_mask.shape[0],
        cleaned_mask.shape[1],
        cleaned_mask.shape[2],
        mask=cleaned_mask.get_fdata(),
        return_as=np.ndarray,
    )
    n_connected_components, labels = connected_components(adjacency)
    assert n_connected_components == 1

    if clean_mask_path is not None:
        cleaned_mask.to_filename(clean_mask_path)
    return cleaned_mask


def select_confounds(cfds_path, cfds_sel):
    cfd = pd.read_csv(cfds_path, sep="\t")
    cols = []
    # cfd = cfd.drop('censored', axis=1)
    if "-censor" in cfds_sel:
        cols.extend(
            [cc for cc in cfd.columns if ("censor" in cc) and (cc != "censored")]
        )
    if "-strictcensor" in cfds_sel:
        cols.extend([cc for cc in cfd.columns if "censfdplus1_" in cc])
        cols.extend([cc for cc in cfd.columns if "strictfd_" in cc])
        cols.extend([cc for cc in cfd.columns if "strictfdplus1_" in cc])
    if "-cosine" in cfds_sel:
        cols.extend([cc for cc in cfd.columns if "cosine" in cc])
    if "-aroma" in cfds_sel:
        cols.extend([cc for cc in cfd.columns if "aroma" in cc])
    if "-motion" in cfds_sel:
        cols.extend([cc for cc in cfd.columns if ("trans" in cc) or ("rot" in cc)])
    if "-physio" in cfds_sel:
        cols.extend([cc for cc in cfd.columns if "phys" in cc])
    if "-gs" in cfds_sel:
        cols.extend(["global_signal"])
    if "-gs+" in cfds_sel:
        cols.extend([cc for cc in cfd.columns if "global_signal" in cc])
    if "-dummy" in cfds_sel:
        cols.extend([cc for cc in cfd.columns if "non_steady_state_outlier" in cc])
    cfds = cfd.loc[:, cols].copy()
    return cfds


def idxs_to_mask(idxs, mask_path):
    mask = nl.image.load_img(mask_path)
    mask_dat = mask.get_fdata()
    null_mask = np.zeros(mask.shape)
    idxs_split = np.array(list((zip(*idxs))))
    null_mask[idxs_split[0], idxs_split[1], idxs_split[2]] = 1
    return null_mask


def idxs_to_zoomed(idxs, mask_path):
    mask = nl.image.load_img(mask_path)
    mask_dat = mask.get_fdata()
    null_mask = np.zeros(mask.shape)
    idxs_split = np.array(list((zip(*idxs))))
    null_mask[idxs_split[0], idxs_split[1], idxs_split[2]] = 1
    maskxs, maskys, maskzs = np.where(mask_dat != 0)
    minx = maskxs.min()
    miny = maskys.min()
    minz = maskzs.min()
    maxx = maskxs.max() + 1
    maxy = maskys.max() + 1
    maxz = maskzs.max() + 1
    clust = null_mask[minx:maxx, miny:maxy, minz:maxz] != 0
    assert (clust != 0).sum() == len(idxs)

    return clust


def idxs_to_flat(idxs, mask_path):
    mask = nl.image.load_img(mask_path)
    mask_dat = mask.get_fdata()
    null_mask = np.zeros(mask.shape)
    idxs_split = np.array(list((zip(*idxs))))
    null_mask[idxs_split[0], idxs_split[1], idxs_split[2]] = 1
    assert len(idxs) == null_mask.sum()
    flat_clust = null_mask[mask_dat != 0]
    assert len(flat_clust) == (mask_dat != 0).sum()
    return flat_clust



from click.testing import CliRunner
from contarg.cli.cli import contarg
from pkg_resources import resource_filename
from pathlib import Path
import nilearn as nl
from nilearn import image, masking
import numpy as np
import pandas as pd


def test_single_subject():
    # get paths
    data_path = Path(resource_filename("contarg", "test/data"))
    bids_dir = data_path / "ds002330"
    derivatives_dir = data_path / "derivatives"
    database_file = data_path / "pybids_0.15.2_db"
    runner = CliRunner()
    result = runner.invoke(
        contarg,
        [
            "hierarchical",
            "run",
            f"--bids-dir={bids_dir}",
            f"--derivatives-dir={derivatives_dir}",
            f"--database-file={database_file}",
            "--run-name=testing1sub",
            "--space=T1w",
            "--smoothing-fwhm=3",
            "--ndummy=5",
            "--tr=1.9",
            "--subject=02",
            "--run=1",
            "--njobs=2",
        ],
    )
    assert result.exit_code == 0
    output_dir = (
        derivatives_dir / "contarg" / "hierarchical" / "testing1sub" / "sub-02" / "func"
    )
    reference_dir = (
        derivatives_dir
        / "contarg"
        / "hierarchical"
        / "testing1sub_ref"
        / "sub-02"
        / "func"
    )
    assert reference_dir.exists()
    assert output_dir.exists()
    niftis = sorted(reference_dir.glob("*.nii.gz"))
    tsvs = sorted(reference_dir.glob("*.tsv"))
    for nii in niftis:
        out_nii = output_dir / nii.name
        out_img = nl.image.load_img(out_nii)
        out_dat = out_img.get_fdata()
        ref_dat = nl.image.load_img(nii).get_fdata()
        assert np.allclose(out_dat, ref_dat)
    for tsv in tsvs:
        out_tsv = output_dir / tsv.name
        out_df = pd.read_csv(out_tsv, sep="\t")
        ref_df = pd.read_csv(tsv, sep="\t")
        assert out_df.equals(ref_df)


def test_multi_subject():
    # get paths
    data_path = Path(resource_filename("contarg", "test/data"))
    bids_dir = data_path / "ds002330"
    derivatives_dir = data_path / "derivatives"
    database_file = data_path / "pybids_0.15.2_db"
    runner = CliRunner()
    result = runner.invoke(
        contarg,
        [
            "hierarchical",
            "run",
            f"--bids-dir={bids_dir}",
            f"--derivatives-dir={derivatives_dir}",
            f"--database-file={database_file}",
            "--run-name=testing2subs",
            "--space=T1w",
            "--smoothing-fwhm=3",
            "--ndummy=5",
            "--tr=1.9",
            "--njobs=4",
        ],
    )
    assert result.exit_code == 0
    output_dir = (
        derivatives_dir
        / "contarg"
        / "hierarchical"
        / "testing2subs"
        / "sub-02"
        / "func"
    )
    reference_dir = (
        derivatives_dir
        / "contarg"
        / "hierarchical"
        / "testing2subs_ref"
        / "sub-02"
        / "func"
    )
    assert reference_dir.exists()
    assert output_dir.exists()
    niftis = sorted(reference_dir.glob("*.nii.gz"))
    tsvs = sorted(reference_dir.glob("*.tsv"))
    for nii in niftis:
        out_nii = output_dir / nii.name
        out_img = nl.image.load_img(out_nii)
        out_dat = out_img.get_fdata()
        ref_dat = nl.image.load_img(nii).get_fdata()
        assert np.allclose(out_dat, ref_dat)
    for tsv in tsvs:
        out_tsv = output_dir / tsv.name
        out_df = pd.read_csv(out_tsv, sep="\t")
        ref_df = pd.read_csv(tsv, sep="\t")
        assert out_df.equals(ref_df)
    output_dir = (
        derivatives_dir
        / "contarg"
        / "hierarchical"
        / "testing2subs"
        / "sub-03"
        / "func"
    )
    reference_dir = (
        derivatives_dir
        / "contarg"
        / "hierarchical"
        / "testing2subs_ref"
        / "sub-03"
        / "func"
    )
    assert reference_dir.exists()
    assert output_dir.exists()
    niftis = sorted(reference_dir.glob("*.nii.gz"))
    tsvs = sorted(reference_dir.glob("*.tsv"))
    for nii in niftis:
        out_nii = output_dir / nii.name
        out_img = nl.image.load_img(out_nii)
        out_dat = out_img.get_fdata()
        ref_dat = nl.image.load_img(nii).get_fdata()
        assert np.allclose(out_dat, ref_dat)
    for tsv in tsvs:
        out_tsv = output_dir / tsv.name
        out_df = pd.read_csv(out_tsv, sep="\t")
        ref_df = pd.read_csv(tsv, sep="\t")
        assert out_df.equals(ref_df)


if __name__ == "__main__":
    test_single_subject()
    test_multi_subject()



from pathlib import Path
from bids import BIDSLayout  # pip version = 0.15.2
import pandas as pd
import nilearn as nl
from nilearn import image, masking
from joblib import Parallel, delayed
import click
from pkg_resources import resource_filename
from contarg.utils import make_path, transform_mask_to_t1w, clean_mask
from contarg.hierarchical import (
    find_target,
    custom_metric,
    rank_clusters,
    get_com_in_mm,
    get_clust_image,
    get_vox_ref_corr,
)


@click.group()
def contarg():
    pass


@contarg.group()
def hierarchical():
    pass


@hierarchical.command()
@click.option("--bids-dir", type=click.Path(exists=True), help="Path to bids root.")
@click.option(
    "--derivatives-dir",
    type=click.Path(exists=True),
    help="Path to derivatives directory with fMRIPrep output.",
)
@click.option(
    "--database-file",
    type=click.Path(),
    help="Path to pybids database file (expects version 0.15.2), "
    "if one does not exist here, it will be created.",
)
@click.option(
    "--run-name",
    type=str,
    default=None,
    help="Name of clustering run. If provided,"
    "output will be placed in derivatives_dir/contarg/hierarchical/run_name. Otherwise,"
    "output will be in derivatives_dir/contarg/hierarchical",
)
@click.option(
    "--stimroi-name",
    type=str,
    default="DLPFCspheres",
    help="Name of roi to which stimulation will be delivered. "
    "Should be one of ['DLPFCspheres', 'BA46sphere'], "
    "or provide the path to the roi in MNI152NLin6Asym space with the stimroi-path option.",
)
@click.option(
    "--stimroi-path",
    type=str,
    default=None,
    help="If providing a custom stim roi, give the path to that ROI here.",
)
@click.option(
    "--refroi-name",
    type=str,
    default="SGCsphere",
    help="Name of roi to whose connectivity is being used to pick a stimulation site. "
    "Should be one of ['SGCsphere', 'bilateralSGCSpheres'], "
    "or provide the path to the roi in MNI152NLin6Asym space with the refroi-path option.",
)
@click.option(
    "--refroi-path",
    type=str,
    default=None,
    help="If providing a custom ref roi, give the path to that ROI here.",
)
@click.option(
    "--space",
    type=click.Choice(["T1w", "MNI152NLin6Asym"]),
    default="T1w",
    show_default=True,
    help="Which space should results be output in. "
    "T1w for directing treatment; "
    "MNI152NLin6Asym for comparing reliabilities across individuals.",
)
@click.option(
    "--smoothing-fwhm",
    type=float,
    help="FWHM in mm of Gaussian blur to apply to functional",
)
@click.option(
    "--ndummy",
    "n_dummy",
    type=int,
    help="Number of dummy scans at the beginning of the functional time series",
)
@click.option(
    "--tr", "t_r", type=float, help="Repetition time of the functional time series"
)
@click.option(
    "--linkage",
    type=click.Choice(["single", "complete", "average"]),
    default="average",
    show_default=True,
    help="Linkage type for clustering",
)
@click.option(
    "--adjacency/--no-adjacency",
    default=False,
    show_default=True,
    help="Whether or not to constrain clustering by face adjacency",
)
@click.option(
    "--distance-threshold",
    type=float,
    default=0.5,
    show_default=True,
    help="Minimum spearman correlation required to cluster",
)
@click.option(
    "--nvox_weight",
    type=float,
    default=1.0,
    show_default=True,
    help="Weight given to number of voxels when selecting a target cluster",
)
@click.option(
    "--concentration-weight",
    type=float,
    default=1.0,
    show_default=True,
    help="Weight given to concentration when selecting a target cluster",
)
@click.option(
    "--nrc-weight",
    "net_reference_correlation_weight",
    type=float,
    default=1.0,
    show_default=True,
    help="Weight given to net reference correlation when selecting a target cluster",
)
@click.option(
    "--subject",
    type=str,
    default=None,
    help="Subject from dataset to generate target(s) for.",
)
@click.option(
    "--session",
    type=str,
    default=None,
    help="Session from dataset to generate target(s) for.",
)
@click.option(
    "--run", type=str, default=None, help="Run from dataset to generate target(s) for."
)
@click.option(
    "--echo",
    type=str,
    default=None,
    help="Echo from dataset to generate target(s) for.",
)
@click.option(
    "--njobs",
    type=int,
    default=1,
    show_default=True,
    help="Number of jobs to run in parallel to find targets",
)
def run(
    bids_dir,
    derivatives_dir,
    database_file,
    run_name,
    stimroi_name,
    stimroi_path,
    refroi_name,
    refroi_path,
    space,
    smoothing_fwhm,
    n_dummy,
    t_r,
    linkage,
    adjacency,
    distance_threshold,
    nvox_weight,
    concentration_weight,
    net_reference_correlation_weight,
    subject,
    session,
    run,
    echo,
    njobs,
):
    bids_dir = Path(bids_dir)
    derivatives_dir = Path(derivatives_dir)
    database_path = Path(database_file)
    roi_dir = Path(resource_filename("contarg", "data/rois"))
    layout = BIDSLayout(
        bids_dir,
        database_path=database_path,
        derivatives=derivatives_dir / "fmriprep",
    )
    if run_name is not None:
        targeting_dir = derivatives_dir / "contarg" / "hierarchical" / run_name
    else:
        targeting_dir = derivatives_dir / "contarg" / "hierarchical"
    targeting_dir.mkdir(parents=True, exist_ok=True)

    if stimroi_name in ["DLPFCspheres", "BA46sphere"]:
        stim_roi_2mm_path = (
            roi_dir / f"{stimroi_name}_space-MNI152NLin6Asym_res-02.nii.gz"
        )
    elif stimroi_path is None:
        raise ValueError(
            f"Custom roi name passed for stimroi, {stimroi_name}, but no path to that roi was provided."
        )
    else:
        stim_roi_2mm_path = stimroi_path

    if refroi_name in ["SGCsphere", "bilateralSGCspheres"]:
        ref_roi_2mm_path = (
            roi_dir / f"{refroi_name}_space-MNI152NLin6Asym_res-02.nii.gz"
        )
    elif refroi_path is None:
        raise ValueError(
            f"Custom roi name passed refroi, {refroi_name}, but no path to that roi was provided."
        )
    else:
        ref_roi_2mm_path = refroi_path

    assert stim_roi_2mm_path.exists()
    assert ref_roi_2mm_path.exists()

    # Getting all the needed input paths
    # build paths df off of bolds info
    get_kwargs = {
        "return_type": "object",
        "task": "rest",
        "desc": "preproc",
        "suffix": "bold",
        "space": "T1w",
        "extension": ".nii.gz",
    }
    if subject is not None:
        get_kwargs["subject"] = subject
    if session is not None:
        get_kwargs["session"] = session
    if run is not None:
        get_kwargs["run"] = run
    if echo is not None:
        get_kwargs["echo"] = echo
    bolds = layout.get(**get_kwargs)
    rest_paths = pd.DataFrame([bb.get_entities() for bb in bolds])
    rest_paths["entities"] = [bb.get_entities() for bb in bolds]
    rest_paths["bold_obj"] = bolds
    rest_paths["bold_path"] = [bb.path for bb in bolds]
    if "session" not in rest_paths.columns:
        rest_paths["session"] = None
    # add boldref
    rest_paths["boldref"] = rest_paths.entities.apply(
        lambda ee: layout.get(
            return_type="file",
            subject=ee["subject"],
            task=ee["task"],
            run=ee["run"],
            datatype="func",
            space="T1w",
            extension=".nii.gz",
            suffix="boldref",
        )
    )
    assert rest_paths.boldref.apply(lambda x: len(x) == 1).all()
    rest_paths["boldref"] = rest_paths.boldref.apply(lambda x: x[0])
    # add brain mask
    rest_paths["brain_mask"] = rest_paths.boldref.str.replace(
        "_boldref", "_desc-brain_mask"
    )
    assert rest_paths.brain_mask.apply(lambda x: Path(x).exists()).all()

    # add t1w path
    rest_paths["T1w"] = rest_paths.entities.apply(
        lambda ee: layout.get(
            return_type="file",
            subject=ee["subject"],
            desc="preproc",
            datatype="anat",
            extension=".nii.gz",
            suffix="T1w",
            space=None,
        )
    )
    assert rest_paths.T1w.apply(lambda x: len(x) == 1).all()
    rest_paths["T1w"] = rest_paths.T1w.apply(lambda x: x[0])

    # add mnito t1w path
    rest_paths["mnitoT1w"] = rest_paths.entities.apply(
        lambda ee: layout.get(
            return_type="file",
            subject=ee["subject"],
            datatype="anat",
            extension=".h5",
            suffix="xfm",
            to="T1w",
            **{"from": "MNI152NLin2009cAsym"},
        )
    )
    assert rest_paths.mnitoT1w.apply(lambda x: len(x) == 1).all()
    rest_paths["mnitoT1w"] = rest_paths.mnitoT1w.apply(lambda x: x[0])

    # add confounds path
    rest_paths["confounds"] = rest_paths.entities.apply(
        lambda ee: layout.get(
            return_type="file",
            subject=ee["subject"],
            task=ee["task"],
            run=ee["run"],
            datatype="func",
            extension=".tsv",
            suffix="timeseries",
            desc="confounds",
        )
    )
    assert rest_paths.confounds.apply(lambda x: len(x) == 1).all()
    rest_paths["confounds"] = rest_paths.confounds.apply(lambda x: x[0])

    if space == "T1w":
        # set up paths to transform stim and target rois back to subject space
        boldmask_pattern = "sub-{subject}/[ses-{session}/]func/sub-{subject}_[ses-{session}_]task-{task}_run-{run}_[echo-{echo}_]atlas-{atlas}_space-{space}_desc-{desc}_{suffix}.{extension}"
        stimmask_updates = {
            "desc": stimroi_name,
            "suffix": "mask",
            "space": space,
            "atlas": "Coords",
        }
        rest_paths[f"stim_mask"] = rest_paths.entities.apply(
            lambda x: make_path(
                x,
                stimmask_updates,
                boldmask_pattern,
                targeting_dir,
                layout.build_path,
                check_exist=False,
                check_parent=False,
                make_parent=True,
            )
        )
        rest_paths["stim_mask_exists"] = rest_paths.stim_mask.apply(
            lambda x: x.exists()
        )

        refmask_updates = {
            "desc": refroi_name,
            "suffix": "mask",
            "space": space,
            "atlas": "Coords",
        }
        rest_paths[f"ref_mask"] = rest_paths.entities.apply(
            lambda x: make_path(
                x,
                refmask_updates,
                boldmask_pattern,
                targeting_dir,
                layout.build_path,
                check_exist=False,
                check_parent=False,
                make_parent=True,
            )
        )
        rest_paths["ref_mask_exists"] = rest_paths.ref_mask.apply(lambda x: x.exists())

        # set up paths for masks with small connected components dropped
        boldmask_pattern = "sub-{subject}/[ses-{session}/]func/sub-{subject}_[ses-{session}_]task-{task}_run-{run}_[echo-{echo}_]atlas-{atlas}_space-{space}_desc-{desc}_{suffix}.{extension}"
        clnstimmask_updates = {
            "desc": stimroi_name + "Clean",
            "suffix": "mask",
            "space": space,
            "atlas": "Coords",
        }
        rest_paths[f"clnstim_mask"] = rest_paths.entities.apply(
            lambda x: make_path(
                x,
                clnstimmask_updates,
                boldmask_pattern,
                targeting_dir,
                layout.build_path,
                check_exist=False,
                check_parent=False,
                make_parent=True,
            )
        )
        rest_paths["clnstim_mask_exists"] = rest_paths.clnstim_mask.apply(
            lambda x: x.exists()
        )

        clnrefmask_updates = {
            "desc": refroi_name + "Clean",
            "suffix": "mask",
            "space": space,
            "atlas": "Coords",
        }
        rest_paths[f"clnref_mask"] = rest_paths.entities.apply(
            lambda x: make_path(
                x,
                clnrefmask_updates,
                boldmask_pattern,
                targeting_dir,
                layout.build_path,
                check_exist=False,
                check_parent=False,
                make_parent=True,
            )
        )
        rest_paths["clnref_mask_exists"] = False

    else:
        # TODO: clean this up for custom ROIs
        clnref_roi_2mm_path = (
            roi_dir / f"{ref_roi_name + 'masked'}_space-MNI152NLin6Asym_res-02.nii.gz"
        )
        rest_paths[f"stim_mask"] = stim_roi_2mm_path
        rest_paths["stim_mask_exists"] = True
        rest_paths["ref_mask"] = ref_roi_2mm_path
        rest_paths["ref_mask_exists"] = True
        rest_paths[f"clnstim_mask"] = stim_roi_2mm_path
        rest_paths["clnstim_mask_exists"] = True
        rest_paths["clnref_mask"] = clnref_roi_2mm_path
        rest_paths["clnref_mask_exists"] = True

    if adjacency:
        desc = f"dt{distance_threshold}.ad.custaff"
    else:
        desc = f"dt{distance_threshold}.custaff"

    clust_updates = {
        "desc": desc,
        "suffix": "targetclust",
        "space": space,
        "atlas": "Coords",
        "extension": ".pkl.gz",
    }
    rest_paths[f"{desc}_targout"] = rest_paths.entities.apply(
        lambda x: make_path(
            x,
            clust_updates,
            boldmask_pattern,
            targeting_dir,
            layout.build_path,
            check_exist=False,
            check_parent=False,
            make_parent=True,
        )
    )
    rest_paths[f"{desc}_targout_exists"] = rest_paths[f"{desc}_targout"].apply(
        lambda x: x.exists()
    )

    clust_updates = {
        "desc": desc,
        "suffix": "targetclust",
        "space": space,
        "atlas": "Coords",
        "extension": ".nii.gz",
    }
    rest_paths[f"{desc}_targout_cluster"] = rest_paths.entities.apply(
        lambda x: make_path(
            x,
            clust_updates,
            boldmask_pattern,
            targeting_dir,
            layout.build_path,
            check_exist=False,
            check_parent=False,
            make_parent=True,
        )
    )
    rest_paths[f"{desc}_targout_cluster_exists"] = rest_paths[
        f"{desc}_targout_cluster"
    ].apply(lambda x: x.exists())

    clust_updates = {
        "desc": desc,
        "suffix": "targetclustinfo",
        "space": space,
        "atlas": "Coords",
        "extension": ".tsv",
    }
    rest_paths[f"{desc}_targout_cluster_info"] = rest_paths.entities.apply(
        lambda x: make_path(
            x,
            clust_updates,
            boldmask_pattern,
            targeting_dir,
            layout.build_path,
            check_exist=False,
            check_parent=False,
            make_parent=True,
        )
    )
    rest_paths[f"{desc}_targout_cluster_info_exists"] = rest_paths[
        f"{desc}_targout_cluster_info"
    ].apply(lambda x: x.exists())

    clust_updates = {
        "desc": desc,
        "suffix": "targetcoords",
        "space": space,
        "atlas": "Coords",
        "extension": ".tsv",
    }
    rest_paths[f"{desc}_targout_coordinates"] = rest_paths.entities.apply(
        lambda x: make_path(
            x,
            clust_updates,
            boldmask_pattern,
            targeting_dir,
            layout.build_path,
            check_exist=False,
            check_parent=False,
            make_parent=True,
        )
    )
    rest_paths[f"{desc}_targout_coordinates_exists"] = rest_paths[
        f"{desc}_targout_coordinates"
    ].apply(lambda x: x.exists())

    clust_updates = {
        "desc": desc,
        "suffix": "net_reference_correlation",
        "space": space,
        "atlas": "Coords",
        "extension": ".nii.gz",
    }
    rest_paths[f"{desc}_net_reference_correlation"] = rest_paths.entities.apply(
        lambda x: make_path(
            x,
            clust_updates,
            boldmask_pattern,
            targeting_dir,
            layout.build_path,
            check_exist=False,
            check_parent=False,
            make_parent=True,
        )
    )
    rest_paths[f"{desc}_net_reference_correlation_exists"] = rest_paths[
        f"{desc}_net_reference_correlation"
    ].apply(lambda x: x.exists())

    # transform all the masks that don't exist
    rest_paths.loc[~rest_paths.stim_mask_exists].apply(
        lambda x: transform_mask_to_t1w(
            x, inmask=stim_roi_2mm_path.as_posix(), outmask_col="stim_mask"
        ),
        axis=1,
    )
    rest_paths["stim_mask_exists"] = rest_paths.stim_mask.apply(lambda x: x.exists())
    assert rest_paths.stim_mask_exists.all()

    rest_paths.loc[~rest_paths.ref_mask_exists].apply(
        lambda x: transform_mask_to_t1w(
            x, inmask=ref_roi_2mm_path.as_posix(), outmask_col="ref_mask"
        ),
        axis=1,
    )
    rest_paths["ref_mask_exists"] = rest_paths.ref_mask.apply(lambda x: x.exists())
    assert rest_paths.ref_mask_exists.all()

    # clean up transformed masks by dropping disconnected bits
    rest_paths.loc[~rest_paths.clnstim_mask_exists].apply(
        lambda row: clean_mask(
            nl.image.load_img(row.stim_mask),
            nl.image.load_img(row.brain_mask),
            max_drop_frac=0.1,
            clean_mask_path=row.clnstim_mask,
            error="skip",
        ),
        axis=1,
    )
    rest_paths["clnstim_mask_exists"] = rest_paths.clnstim_mask.apply(
        lambda x: x.exists()
    )
    if not rest_paths.clnstim_mask_exists.all():
        raise ValueError("Could not create a clean mask")

    rest_paths.loc[~rest_paths.clnref_mask_exists].apply(
        lambda row: clean_mask(
            nl.image.load_img(row.ref_mask),
            nl.image.load_img(row.brain_mask),
            max_drop_frac=0.1,
            clean_mask_path=row.clnref_mask,
            error="skip",
        ),
        axis=1,
    )
    rest_paths["clnref_mask_exists"] = rest_paths.clnref_mask.apply(
        lambda x: x.exists()
    )
    if not rest_paths.clnref_mask_exists.all():
        raise ValueError("Could not create a clean mask")

    # now that we've got all the paths set, lets start finding targets
    jobs = []
    for ix, row in rest_paths.iterrows():
        jobs.append(
            delayed(find_target)(
                row,
                "bold_path",
                f"{desc}_targout",
                n_dummy=n_dummy,
                t_r=t_r,
                distance_threshold=distance_threshold,
                adjacency=adjacency,
                smoothing_fwhm=smoothing_fwhm,
                confound_selectors=["-gs", "-motion", "-dummy"],
                metric=custom_metric,
                linkage=linkage,
                write_pkl=True,
                return_df=True,
            )
        )

    r = Parallel(n_jobs=njobs, verbose=10)(jobs)
    targouts = pd.concat(r)

    # merge some paths in to the targets dataframe
    rp_cols = [
        "subject",
        "session",
        "run",
        "clnstim_mask",
        "clnref_mask",
        "bold_path",
        "brain_mask",
        "confounds",
        f"{desc}_targout",
        f"{desc}_targout_exists",
        f"{desc}_targout_cluster",
        f"{desc}_targout_cluster_exists",
        f"{desc}_targout_cluster_info",
        f"{desc}_targout_cluster_info_exists",
        f"{desc}_targout_coordinates",
        f"{desc}_targout_coordinates_exists",
        f"{desc}_net_reference_correlation",
        f"{desc}_net_reference_correlation_exists",
    ]
    tmp = targouts.merge(
        rest_paths.loc[:, rp_cols],
        how="left",
        on=["subject", "session", "run"],
        indicator=True,
    )
    assert (tmp._merge != "both").sum() == 0
    targouts_mp = targouts.merge(
        rest_paths.loc[:, rp_cols], how="left", on=["subject", "session", "run"]
    )

    # rank clusters
    targouts_sel = rank_clusters(
        targouts_mp, nvox_weight, concentration_weight, net_reference_correlation_weight
    )

    # write out results
    for ixs, df in targouts_sel.groupby(["run", "session", "subject"], dropna=False):
        target_row = df.query("overall_rank == 1").iloc[0]
        target_cluster = get_clust_image(target_row)
        target_cluster.to_filename(target_row[f"{desc}_targout_cluster"])
        df.to_csv(target_row[f"{desc}_targout_cluster_info"], index=None, sep="\t")
        target_coords = get_com_in_mm(target_row)
        target_coords_df = pd.DataFrame(data=[target_coords], columns=["x", "y", "z"])
        target_coords_df.to_csv(
            target_row[f"{desc}_targout_cluster_info"], index=None, sep="\t"
        )

        vox_ref_corr_img = get_vox_ref_corr(
            target_row,
            distance_threshold,
            adjacency,
            smoothing_fwhm,
            ["-gs", "-motion", "-dummy"],
            custom_metric,
            linkage,
            t_r,
            n_dummy,
        )
        vox_ref_corr_img.to_filename(target_row[f"{desc}_net_reference_correlation"])
